#include <iostream>
#include <bitset>
#include <string>

const unsigned maxArgs=10;
    
typedef std::bitset<maxArgs> ArgBits;

template<unsigned numBits>
bool increment(std::bitset<numBits>& bits)
{
    for(unsigned pos=0;pos<numBits;++pos)
    {
	const bool wasSet=bits[pos];
	bits.flip(pos);
	if(!wasSet)
	{
	    return true;
	}
    }
    return false;
}

template<unsigned numBits>
unsigned countSet(const std::bitset<numBits>& argBits)
{
    unsigned argCount=0;
    for(unsigned index=0;index<numBits;++index)
    {
	if(!argBits[index]) continue;
	
	++argCount;
    }
    return argCount;
}

template<unsigned numBits>
unsigned maxSet(const std::bitset<numBits>& argBits)
{
    unsigned argCount=~0;
    for(unsigned index=0;index<numBits;++index)
    {
	if(!argBits[index]) continue;
	
	argCount=index;
    }
    return argCount+1;
}

void repeatWrite(std::ostream& os,unsigned count,const std::string& s)
{
    for(;count;--count)
    {
	os<<s;
    }
}

template<unsigned numBits,typename Func>
void writeCombination(std::ostream& os,const std::bitset<numBits>& argBits,unsigned maxIndex,Func entryFunc,const std::string& betweenString)
{
    for(unsigned index=0;index<maxIndex && index<numBits;++index)
    {
	if(index)
	{
	    os<<betweenString;
	}
	
	entryFunc(os,index,argBits[index]);
    }
}

void writeArgParam(std::ostream& os,unsigned index,bool isSet)
{
    if(!isSet)
    {
	os<<"const ";
    }
    os<<"A"<<index+1<<"& a"<<index+1;
}

void writeNCEntry(std::ostream& os,unsigned index,bool isSet)
{
    if(isSet)
    {
	os<<"ArgSets::NCRefPosEntry<"<<index+1<<",";
    }
}

void writeArgSetEntry(std::ostream& os,unsigned index,bool isSet)
{
    if(isSet)
    {
	os<<"typename Holder::RefType<";
    }
    os<<"A"<<index+1;
    if(isSet)
    {
	os<<">::Type";
    }
}

void writeArg(std::ostream& os,unsigned index,bool isSet)
{
    os<<"a"<<index+1;
}

template<unsigned>
struct Index
{};

int main()
{
    ArgBits argBits(0);

    std::cout<<"#ifndef FUNCCALLOPIMPL_HH\n"
	"#define FUNCCALLOPIMPL_HH\n"
	"\n"
	"namespace FuncImpl\n"
	"{\n";
    
    do
    {
	const unsigned argCount=countSet(argBits);
	const unsigned maxCount=maxSet(argBits);

	if(argCount)
	{
	    std::cout<<"template<typename FuncCaller>\n"
		"struct FunctionCallOperatorImpl<FuncCaller,";
	
	    writeCombination(std::cout,argBits,maxArgs,writeNCEntry,"");
	    std::cout<<"ArgSets::NCRefEndOfList>";
	    repeatWrite(std::cout,argCount," >");
	}
	else
	{
	    std::cout<<"template<typename FuncCaller,typename NCRefPosSet>\n"
		"struct FunctionCallOperatorImpl";
	}
	
	std::cout<<"\n"
	    "{\n";

	for(unsigned numArgs=maxCount;numArgs<=maxArgs;++numArgs)
	{
	    if(numArgs)
	    {
		std::cout<<"    template<";
		for(unsigned argNum=1;argNum<numArgs;++argNum)
		{
		    std::cout<<"typename A"<<argNum<<",";
		}
		std::cout<<"typename A"<<numArgs<<">\n";
	    }
	    
	    std::cout<<"    typename FuncCaller::template Helper<ArgSets::ArgSetRef"<<numArgs;
	    if(numArgs)
	    {
		std::cout<<"<";
		writeCombination(std::cout,argBits,numArgs,writeArgSetEntry,",");
		std::cout<<" >";
	    }
	    
	    std::cout<<" >::ReturnType operator()(";
	    writeCombination(std::cout,argBits,numArgs,writeArgParam,",");
	    std::cout<<") const\n"
		"    {\n"
		"        typedef ArgSets::ArgSetRef"<<numArgs;
	    if(numArgs)
	    {
		std::cout<<"<";
		writeCombination(std::cout,argBits,numArgs,writeArgSetEntry,",");
		std::cout<<" >";
	    }
	    std::cout<<" AS;\n"
		"        return static_cast<typename FuncCaller::DerivedType>(*this).call(AS(";
	    writeCombination(std::cout,argBits,numArgs,writeArg,",");
	    std::cout<<"));\n"
		"    }\n";
	}
	

	std::cout<<"};\n"<<std::endl;
    }
    while(increment(argBits));

    std::cout<<"}\n"
	"#endif"<<std::endl;
}
