#include <iostream>
#include <typeinfo>
#include <functional>
#include <algorithm>
#include <vector>
#include <string>
#include <math.h>
#include <cassert>

#include "compose.hh"
#include "argset_transform.hh"

using namespace Compose;

struct F
{
    double operator()(double i,double d) const
    {
	std::cout<<"i="<<i<<", d="<<d<<std::endl;
	return i*d*10;
    }
    typedef double result_type;
} f;

struct G
{
    double operator()(double i) const
    {
	std::cout<<"i="<<i<<std::endl;
	return i*10;
    }
    typedef double result_type;
} g;

struct H
{
    double operator()(double i1,double i2,double i3) const
    {
	std::cout<<"i1="<<i1<<",i2="<<i2<<",i3="<<i3<<std::endl;
	return i1*i2*i3;
    }
    typedef double result_type;
} h;

struct J
{
    double operator()(double i1,double i2,double i3,double i4) const
    {
	std::cout<<"i1="<<i1<<",i2="<<i2<<",i3="<<i3<<",i4="<<i4<<std::endl;
	return i1*i2*i3*i4;
    }
    typedef double result_type;
} j;

struct RefReturn
{
    int& operator()(int& i) const
    {
	std::cout<<"ref i="<<i<<std::endl;
	return i;
    }
    
} refReturn;



void write(double d)
{
    std::cout<<d<<std::endl;
}


template<typename Func>
void repeatCall(Func f)
{
    std::cout<<"Calling"<<std::endl;
    f(1,2,3,4,5,6,7,8,9,10);
    std::cout<<"Calling"<<std::endl;
    f(1,2,3,4,5,6,7,8,9,10);
    std::cout<<"Calling"<<std::endl;
    f(1,2,3,4,5,6,7,8,9,10);
}

REGISTER_EXPR_TYPE(Holder::RefHolder<int>,1000)
REGISTER_EXPR_TYPE(Holder::RefHolder<double>,1002)
REGISTER_EXPR_TYPE(std::string,1003)



    struct Inc:
    std::unary_function<int&,void>
{
    void operator()(int& y) const
    {
	std::cout<<"y="<<y<<std::endl;
	++y;
    }
    typedef void result_type;
} inc;


struct Myfunc
{
    void operator()() const
    {
	std::cout<<"void MyFunc()"<<std::endl;
    }
    typedef void result_type;
} myfunc;

namespace FuncImpl
{
    template<>
    struct HasResultType<Myfunc>
    {
	static const bool result=true;
    };
}


int normalFunc(int i)
{
    std::cout<<"normalFunc,i="<<i<<std::endl;
    return 10*i;
}

int noArgsFunc()
{
    std::cout<<"noArgsFunc called"<<std::endl;
    return 23;
}

double updateFunc(double& i)
{
    std::cout<<"updateFunc called,i="<<i<<std::endl;
    return i*=10;
}

template<typename T>
void cvtest(T&)
{
    std::cout<<"cvtest: "<<typeid(T).name()<<std::endl;
}

template<typename T>
void cvtest(const T&)
{
    std::cout<<"cvtest: const "<<typeid(T).name()<<std::endl;
}

template<typename T>
void cvtest(volatile T&)
{
    std::cout<<"cvtest: volatile "<<typeid(T).name()<<std::endl;
}

template<typename T>
void cvtest(const volatile T&)
{
    std::cout<<"cvtest: const volatile "<<typeid(T).name()<<std::endl;
}


int main()
{
    std::cout<<"simple"<<std::endl;
    assert(compose(g,4.5)()==(4.5*10.0));

    assert(ArgSets::MinArgs<ArgSets::ArgSet0>::result==0);
    assert(ArgSets::MinArgs<ArgSets::ArgSet1<double> >::result==0);
    assert(ArgSets::MinArgs<ArgSets::ArgSet1<FuncImpl::ParamType<1> > >::result==1);
    unsigned res=ArgSets::MinArgs<ArgSets::ArgSet2<FuncImpl::ParamType<1>,FuncImpl::ParamType<2> > >::result;
    assert(res==2);
    assert(ArgSets::NumArgs<FuncImpl::ParamType<1> >::result==1);
    res=CT::Max<unsigned,ArgSets::NumArgs<FuncImpl::ParamType<1> >::result,ArgSets::MinArgs<ArgSets::ArgSet0>::result>::result;
    assert(res==1);

    typedef ArgSets::ArgSet1<FuncImpl::ParamType<1> > AS;
    assert(std::string(typeid(ArgSets::ArgSet1<FuncImpl::ParamType<1> >::Arg).name())==typeid(AS::Arg).name());
    assert(std::string(typeid(ArgSets::ArgSet1<FuncImpl::ParamType<1> >::Base).name())==typeid(AS::Base).name());

    typedef AS::Arg Arg;
    typedef AS::Base Base;
    
    assert(ArgSets::NumArgs<AS::Arg>::result==1);
    assert(ArgSets::MinArgs<AS::Base>::result==0);
    res=CT::Max<unsigned,ArgSets::NumArgs<AS::Arg>::result,
	ArgSets::MinArgs<AS::Base>::result>::result;
    assert(res==1);

    std::cout<<"simple"<<std::endl;
    assert(compose(g,_1)(3.7)==3.7*10);
    
    std::cout<<"simple"<<std::endl;
    assert(compose(f,3,_1)(4.5)==3.0*4.5*10);
    std::cout<<"simple"<<std::endl;
    assert(compose(f,_1,4.7)(1.2)==1.2*4.7*10);
    std::cout<<"simple"<<std::endl;
    assert(compose(f,_2,_1)(1.3,1.4)==1.3*1.4*10);
    std::cout<<"simple"<<std::endl;
    assert(compose(f,1.1,1.2)()==1.1*1.2*10);
    std::cout<<"simple"<<std::endl;
    assert(compose(f,1.1,1.2)(1.4)==1.1*1.2*10);
    std::cout<<"simple"<<std::endl;
    assert(compose(g,9.9)()==9.9*10);
    std::cout<<"simple"<<std::endl;
    assert(compose(g,9.9)(1.7)==9.9*10);
    std::cout<<"compose"<<std::endl;
    assert(compose(g,compose(g,_2))(1,2)==2.0*10*10);
    std::cout<<"compose compose"<<std::endl;
    assert(compose(g,compose(g,compose(g,_2)))(1,2)==2.0*10*10*10);
    std::cout<<"compose compose"<<std::endl;
    compose(g,compose(f,compose(g,_1),_1))(1);
    std::cout<<"compose compose"<<std::endl;
    compose(g,compose(f,compose(g,_1),_2))(1,2);
    std::cout<<"compose compose"<<std::endl;
    compose(g,compose(f,compose(g,_2),_1))(1,2);
    std::cout<<"compose compose"<<std::endl;
    compose(g,compose(f,_1,compose(g,_1)))(1);
    std::cout<<"compose compose"<<std::endl;
    compose(g,compose(f,_2,compose(g,_1)))(1,2);
    std::cout<<"compose compose"<<std::endl;
    compose(g,compose(f,_1,compose(g,_2)))(1,2);
    std::cout<<"compose double compose"<<std::endl;
    compose(g,compose(f,compose(g,_1),compose(g,_2)))(1,2);
    std::cout<<"compose double compose"<<std::endl;
    compose(g,compose(f,compose(g,_1),compose(g,_3)))(1,2,3);

    std::cout<<"extra"<<std::endl;

    compose(f,compose(f,_2,compose(g,compose(f,_1,_2))),_1)(1.9,2.5);
    std::cout<<"extra"<<std::endl;
    compose(f,compose(f,_2,compose(g,_1)),_2)(9.9,1.4);
    std::cout<<"extra"<<std::endl;
    compose(f,compose(f,_1,compose(g,_2)),_2)(9.9,1.4);

    std::cout<<"no args"<<std::endl;
    compose(g,compose(g,1))();

    std::cout<<"some args"<<std::endl;
    compose(g,compose(f,_2,3.5))(1,2);

    std::cout<<"f _ _"<<std::endl;
    compose(g,compose(h,compose(g,_1),_2,_3))(1,2,3);
    std::cout<<"_ f _"<<std::endl;
    compose(g,compose(h,_1,compose(g,_2),_3))(1,2,3);
    std::cout<<"_ _ f"<<std::endl;
    compose(g,compose(h,_1,_2,compose(g,_3)))(1,2,3);

    std::cout<<"f f _"<<std::endl;
    compose(g,compose(h,compose(g,_1),compose(g,_2),_3))(1,2,3);
    std::cout<<"f _ f"<<std::endl;
    compose(g,compose(h,compose(g,_1),_2,compose(g,_3)))(1,2,3);
    std::cout<<"_ f f"<<std::endl;
    compose(g,compose(h,_1,compose(g,_2),compose(g,_3)))(1,2,3);
    std::cout<<"f f f"<<std::endl;
    compose(g,compose(h,compose(g,_1),compose(g,_2),compose(g,_3)))(1,2,3);

    std::cout<<"f _ _ _"<<std::endl;
    compose(g,compose(j,compose(g,_1),_2,_3,_4))(1,2,3,4);
    std::cout<<"_ f _ _"<<std::endl;
    compose(g,compose(j,_1,compose(g,_2),_3,_4))(1,2,3,4);
    std::cout<<"_ _ f _"<<std::endl;
    compose(g,compose(j,_1,_2,compose(g,_3),_4))(1,2,3,4);
    std::cout<<"_ _ _ f"<<std::endl;
    compose(g,compose(j,_1,_2,_3,compose(g,_4)))(1,2,3,4);

    std::cout<<"f f _ _"<<std::endl;
    compose(g,compose(j,compose(g,_1),compose(g,_2),_3,_4))(1,2,3,4);
    std::cout<<"f _ f _"<<std::endl;
    compose(g,compose(j,compose(g,_1),_2,compose(g,_3),_4))(1,2,3,4);
    std::cout<<"f _ _ f"<<std::endl;
    compose(g,compose(j,compose(g,_1),_2,_3,compose(g,_4)))(1,2,3,4);
    std::cout<<"_ f f _"<<std::endl;
    compose(g,compose(j,_1,compose(g,_2),compose(g,_3),_4))(1,2,3,4);
    std::cout<<"_ f _ f"<<std::endl;
    compose(g,compose(j,_1,compose(g,_2),_3,compose(g,_4)))(1,2,3,4);
    std::cout<<"_ _ f f"<<std::endl;
    compose(g,compose(j,_1,_2,compose(g,_3),compose(g,_4)))(1,2,3,4);


    std::cout<<"f f f _"<<std::endl;
    compose(g,compose(j,compose(g,_1),compose(g,_2),compose(g,_3),_4))(1,2,3,4);
    std::cout<<"f f _ f"<<std::endl;
    compose(g,compose(j,compose(g,_1),compose(g,_2),_3,compose(g,_4)))(1,2,3,4);
    std::cout<<"f _ f f"<<std::endl;
    compose(g,compose(j,compose(g,_1),_2,compose(g,_3),compose(g,_4)))(1,2,3,4);

    std::cout<<"f f f f"<<std::endl;
    compose(g,compose(j,compose(g,_1),compose(g,_2),compose(g,_3),compose(g,_4)))(1,2,3,4);

    repeatCall(compose(g,compose(j,compose(g,compose(f,1.5,2.7)),compose(h,_1,_2,_3),_1,compose(g,_2))));

    std::cout<<"call Compose"<<std::endl;
    
    compose(g,_1)(1);
    std::cout<<"call Compose"<<std::endl;
    compose(h,5,_1,compose(g,_2))(1,2);

    int x=1;

    std::cout<<"by ref"<<std::endl;
    compose(inc,byRef(x))();
    repeatCall(compose(inc,byRef(x)));
    
    std::cout<<"x="<<x<<std::endl;

    std::cout<<"Empty args"<<std::endl;
    compose(myfunc)();

    std::cout<<"Plus"<<std::endl;
    compose(g,compose(std::plus<double>(),_1,1))(3);

    compose(normalFunc,compose(normalFunc,_1))(3);

    repeatCall(compose(myfunc));

    int zz=compose(noArgsFunc)();

    std::cout<<"zz="<<zz<<std::endl;
    
    zz=compose(g,compose(j,compose(noArgsFunc),compose(noArgsFunc),compose(noArgsFunc),compose(noArgsFunc)))();
    
    std::cout<<"zz="<<zz<<std::endl;

    zz=compose(g,compose(j,compose(noArgsFunc),compose(noArgsFunc),compose(noArgsFunc),compose(noArgsFunc)))(1,2,3,4);
    
    std::cout<<"zz="<<zz<<std::endl;

    std::vector<double> vd;
    {
	for(unsigned i=0;i<10;++i)
	{
	    vd.push_back(i);
	}
    }

    std::for_each(vd.begin(),vd.end(),
		 compose(write,compose(std::multiplies<double>(),_1,2)));
    std::for_each(vd.begin(),vd.end(),
		 compose(write,compose(std::multiplies<double>(),_1,3)));

    typedef double (*sinCosFunc)(double);
    
    std::for_each(vd.begin(),vd.end(),
		 compose(write,compose((sinCosFunc)sin,_1)));
    std::for_each(vd.begin(),vd.end(),
		 compose(write,compose((sinCosFunc)cos,_1)));
    std::for_each(vd.begin(),vd.end(),
		 compose(write,compose(std::plus<double>(),_1,2)));

    std::cout<<"original sequence"<<std::endl;
    std::for_each(vd.begin(),vd.end(),
		 compose(write,_1));
    std::cout<<"byRef(_1) with std::for_each"<<std::endl;
    std::for_each(vd.begin(),vd.end(),
		 compose(write,compose(updateFunc,byRef(_1))));

    std::cout<<"updated sequence"<<std::endl;
    std::for_each(vd.begin(),vd.end(),
		 compose(write,_1));

    std::cout<<"many args by ref"<<std::endl;
    double x1=1,x2=2,x3=9,x4=10;
    std::cout<<"x1="<<x1<<std::endl;
    std::cout<<"x2="<<x2<<std::endl;
    std::cout<<"x3="<<x3<<std::endl;
    std::cout<<"x4="<<x4<<std::endl;
    compose(j,compose(updateFunc,byRef(_1)),compose(updateFunc,byRef(_2)),compose(updateFunc,byRef(_9)),compose(updateFunc,byRef(_10)))(x1,x2,3,4,5,6,7,8,x3,byRef(x4));
    std::cout<<"x1="<<x1<<std::endl;
    std::cout<<"x2="<<x2<<std::endl;
    std::cout<<"x3="<<x3<<std::endl;
    std::cout<<"x4="<<x4<<std::endl;

    std::cout<<"typeof ref(ref(x))="<<typeid(byRef(byRef(x4))).name()<<std::endl;

    volatile std::string vi="hello";
    const volatile std::string cvi="hello";
    
    EXPR_TYPE(vi) vi2="hello";
    EXPR_TYPE(cvi) cvi2="hello";

    cvtest(vi);
    cvtest(cvi);
    cvtest(vi2);
    cvtest(cvi2);

    // double compose of void-returning func pointer
    compose(compose(write,_1),_1)(23);

    // check references

    int testInt=99;
    
    std::cout<<"testInt="<<testInt<<std::endl;
    int& testIntRef=compose(refReturn,byRef(_1))(testInt);
    ++testIntRef;

    std::cout<<"testInt="<<testInt<<std::endl;
    std::cout<<"testIntRef="<<testIntRef<<std::endl;

    return 0;
}
